@@ -117,6 +117,9 @@ static char sccsid[] = "@(#)login.c	5.40 (Berkeley) 5/9/89";
 #include <sys/syslog.h>
 #include <sys/sysmacros.h>
 #include <netdb.h>
+#ifdef linux
+#  include <linux/major.h>
+#endif
 
 #ifdef TESTING
 #  include "utmp.h"
@@ -130,8 +133,11 @@ static char sccsid[] = "@(#)login.c	5.40 (Berkeley) 5/9/89";
 
 #ifndef linux
 #include <tzfile.h>
+#endif
 #include <lastlog.h>
-#else
+
+#if 0
+/* from before we had a lastlog.h file in linux */
 struct  lastlog
   { long ll_time;
     char ll_line[12];
@@ -164,13 +170,11 @@ char	realm[REALM_SZ];
 int	kerror = KSUCCESS, notickets = 1;
 #endif
 
-#ifndef linux
 #define	TTYGRPNAME	"tty"		/* name of group to own ttys */
-#else
-#  define TTYGRPNAME      "other"
-#  ifndef MAXPATHLEN
-#    define MAXPATHLEN 1024
-#  endif
+/**#  define TTYGRPNAME      "other" **/
+
+#ifndef MAXPATHLEN
+#  define MAXPATHLEN 1024
 #endif
 
 /*
@@ -186,7 +190,7 @@ int     timeout = 60;
 struct	passwd *pwd;
 int	failures;
 char	term[64], *hostname, *username, *tty;
-
+struct  hostent hostaddress;
 char	thishost[100];
 
 #ifndef linux
@@ -265,6 +269,7 @@ main(argc, argv)
 	(void)strncpy(thishost, tbuf, sizeof(thishost)-1);
 	domain = index(tbuf, '.');
 
+	hostname = NULL;
 	fflag = hflag = pflag = 0;
 	passwd_req = 1;
 	while ((ch = getopt(argc, argv, "fh:p")) != EOF)
@@ -284,6 +289,14 @@ main(argc, argv)
 			    strcasecmp(p, domain) == 0)
 				*p = 0;
 			hostname = optarg;
+			{ 
+			    struct hostent *he = gethostbyname(hostname);
+			    if (he) {
+				memcpy(&hostaddress, he, sizeof(hostaddress));
+			    } else {
+				memset(&hostaddress, 0, sizeof(hostaddress));
+			    }
+			}
 			break;
 
 		case 'p':
@@ -357,7 +370,7 @@ main(argc, argv)
 	    tcsetattr(0,TCSAFLUSH,&tt);
 	}
 
-	if (tty = rindex(ttyn, '/'))
+	if ((tty = rindex(ttyn, '/')))
 		++tty;
 	else
 		tty = ttyn;
@@ -375,10 +388,23 @@ main(argc, argv)
 			getloginname();
 		}
 
+               /* Dirty patch to fix a gigantic security hole when using 
+                  yellow pages. This problem should be solved by the
+                  libraries, and not by programs, but this must be fixed
+                  urgently! If the first char of the username is '+', we 
+                  avoid login success.
+                  Feb 95 <alvaro@etsit.upm.es> */
+
+		if (username[0] == '+') {
+		    puts("Illegal username");
+		    badlogin(username);
+		    sleepexit(1);
+		}
+
 		checktty(username, tty);
 
 		(void)strcpy(tbuf, username);
-		if (pwd = getpwnam(username))
+		if ((pwd = getpwnam(username)))
 			salt = pwd->pw_passwd;
 		else
 			salt = "xx";
@@ -566,30 +592,38 @@ main(argc, argv)
 	/* for linux, write entries in utmp and wtmp */
 	{
 		struct utmp ut;
-		char *ttyabbrev;
 		int wtmp;
-		
-		memset((char *)&ut, 0, sizeof(ut));
-		ut.ut_type = USER_PROCESS;
-		ut.ut_pid = getpid();
-		strncpy(ut.ut_line, ttyn + sizeof("/dev/")-1, sizeof(ut.ut_line));
-		ttyabbrev = ttyn + sizeof("/dev/tty") - 1;
-		strncpy(ut.ut_id, ttyabbrev, sizeof(ut.ut_id));
-		(void)time(&ut.ut_time);
-		strncpy(ut.ut_user, username, sizeof(ut.ut_user));
-		
-		/* fill in host and ip-addr fields when we get networking */
-		if (hostname) {
-		    struct hostent *he;
+		struct utmp *utp;
+		pid_t mypid = getpid();
 
-		    strncpy(ut.ut_host, hostname, sizeof(ut.ut_host));
-		    if ((he = gethostbyname(hostname)))
-		      memcpy(&ut.ut_addr, he->h_addr_list[0],
-			     sizeof(ut.ut_addr));
-		}
-		
 		utmpname(_PATH_UTMP);
 		setutent();
+		while ((utp = getutent()) 
+		       && !(utp->ut_pid == mypid)) /* nothing */;
+
+		if (utp) {
+		    memcpy(&ut, utp, sizeof(ut));
+		} else {
+		    /* some gettys/telnetds don't initialize utmp... */
+		    memset(&ut, 0, sizeof(ut));
+		}
+		endutent();
+
+		if (ut.ut_id[0] == 0)
+		  strncpy(ut.ut_id, ttyn + 8, sizeof(ut.ut_id));
+	
+		strncpy(ut.ut_user, username, sizeof(ut.ut_user));
+		strncpy(ut.ut_line, ttyn + 5, sizeof(ut.ut_line));
+		time(&ut.ut_time);
+		ut.ut_type = USER_PROCESS;
+		ut.ut_pid = mypid;
+		if (hostname) {
+		    strncpy(ut.ut_host, hostname, sizeof(ut.ut_host));
+		    if (hostaddress.h_addr_list)
+		      memcpy(&ut.ut_addr, hostaddress.h_addr_list[0],
+			     sizeof(ut.ut_addr));
+		}
+
 		pututline(&ut);
 		endutent();
 		
@@ -600,7 +634,6 @@ main(argc, argv)
 			close(wtmp);
 		}
 	}
-        /* fix_utmp_type_and_user(username, ttyn, LOGIN_PROCESS); */
 #endif
 
 	dolastlog(quietlog);
@@ -615,9 +648,13 @@ main(argc, argv)
 	(void)chown(ttyn, pwd->pw_uid,
 	    (gr = getgrnam(TTYGRPNAME)) ? gr->gr_gid : pwd->pw_gid);
 
-	(void)chmod(ttyn, 0622);
-	(void)setgid(pwd->pw_gid);
+#ifdef USE_TTY_GROUP
+	chmod(ttyn, 0620);
+#else
+	chmod(ttyn, 0600);
+#endif
 
+	setgid(pwd->pw_gid);
 	initgroups(username, pwd->pw_gid);
 
 #ifdef HAVE_QUOTA
@@ -676,10 +713,9 @@ main(argc, argv)
 	(void)setenv("LOGNAME", pwd->pw_name, 1);
 #endif
 
-#ifndef linux
-	if (tty[sizeof("tty")-1] == 'd')
+	if (tty[sizeof("tty")-1] == 'S')
 		syslog(LOG_INFO, "DIALUP %s, %s", tty, pwd->pw_name);
-#endif
+
 	if (pwd->pw_uid == 0)
 		if (hostname)
 			syslog(LOG_NOTICE, "ROOT LOGIN ON %s FROM %s",
@@ -748,16 +784,26 @@ getloginname()
 	register int ch;
 	register char *p;
 	static char nbuf[UT_NAMESIZE + 1];
+	int cnt, cnt2;
 
+	cnt2 = 0;
 	for (;;) {
+	        cnt = 0;
 		(void)printf("\n%s login: ", thishost); fflush(stdout);
 		for (p = nbuf; (ch = getchar()) != '\n'; ) {
 			if (ch == EOF) {
-				badlogin(username);
+				badlogin("EOF");
 				exit(0);
 			}
 			if (p < nbuf + UT_NAMESIZE)
 				*p++ = ch;
+
+			cnt++;
+			if (cnt > UT_NAMESIZE + 20) {
+				fprintf(stderr, "login name much too long.\n");
+				badlogin("NAME too long");
+				exit(0);
+			}
 		}
 		if (p > nbuf)
 			if (nbuf[0] == '-')
@@ -768,6 +814,13 @@ getloginname()
 				username = nbuf;
 				break;
 			}
+
+		cnt2++;
+		if (cnt2 > 50) {
+			fprintf(stderr, "too many bare linefeeds.\n");
+			badlogin("EXCESSIVE linefeeds");
+			exit(0);
+		}
 	}
 }
 
@@ -923,45 +976,105 @@ stypeof(ttyid)
 }
 #endif 
 
+/* return true if tty is a pty. Very linux dependent */
+static int
+isapty(tty)
+     char *tty;
+{
+    char devname[100];
+    struct stat stb;
+
+#ifdef linux
+    strcpy(devname, "/dev/");
+    strncat(devname, tty, 80);
+    if((stat(devname, &stb) >= 0)
+       && major(stb.st_rdev) == TTY_MAJOR
+       && minor(stb.st_rdev) >= 192) {
+	return 1;
+    }
+#endif
+    return 0;
+}
+
+/* match the hostname hn against the pattern pat */
+static int
+hnmatch(hn, pat)
+     char *hn;
+     char *pat;
+{
+    int x1, x2, x3, x4, y1, y2, y3, y4;
+    unsigned long p, mask, a;
+    unsigned char *ha;
+    int n, m;
+
+    if ((hn == NULL) && (strcmp(pat, "localhost") == 0)) return 1;
+    if ((hn == NULL) || hn[0] == 0) return 0;
+
+    if (pat[0] >= '0' && pat[0] <= '9') {
+	/* pattern is an IP QUAD address and a mask x.x.x.x/y.y.y.y */
+	sscanf(pat, "%d.%d.%d.%d/%d.%d.%d.%d", &x1, &x2, &x3, &x4,
+	       &y1, &y2, &y3, &y4);
+	p = (((unsigned long)x1<<24)+((unsigned long)x2<<16)
+	     +((unsigned long)x3<<8)+((unsigned long)x4));
+	mask = (((unsigned long)y1<<24)+((unsigned long)y2<<16)
+		+((unsigned long)y3<<8)+((unsigned long)y4));
+
+	if (!hostaddress.h_addr_list || !hostaddress.h_addr_list[0])
+	  return 0;
+
+	ha = (unsigned char *)hostaddress.h_addr_list[0];
+	a = (((unsigned long)ha[0]<<24)+((unsigned long)ha[1]<<16)
+	     +((unsigned long)ha[2]<<8)+((unsigned long)ha[3]));
+	return ((p & mask) == (a & mask));
+    } else {
+	/* pattern is a suffix of a FQDN */
+	n = strlen(pat);
+	m = strlen(hn);
+	if (n > m) return 0;
+	return (strcasecmp(pat, hn + m - n) == 0);
+    }
+}
+
 void
 checktty(user, tty)
      char *user;
      char *tty;
 {
     FILE *f;
-    char buf[256];
+    char buf[256], defaultbuf[256];
     char *ptr;
-    char devname[50];
-    struct stat stb;
 
     /* no /etc/usertty, default to allow access */
     if(!(f = fopen(_PATH_USERTTY, "r"))) return;
 
+    defaultbuf[0] = 0;
     while(fgets(buf, 255, f)) {
 
 	/* strip comments */
 	for(ptr = buf; ptr < buf + 256; ptr++) 
 	  if(*ptr == '#') *ptr = 0;
 
+	if (buf[0] == '*') {
+	    strncpy(defaultbuf, buf, 256);
+	}
+
 	strtok(buf, " \t");
 	if(strncmp(user, buf, 8) == 0) {
 	    while((ptr = strtok(NULL, "\t\n "))) {
-		if(strncmp(tty, ptr, 10) == 0) {
+		if (strncmp(tty, ptr, 10) == 0) {
+		    fclose(f);
+		    return;
+		}
+#if 0
+		if ((strcmp("PTY", ptr) == 0) && isapty(tty)) {
 		    fclose(f);
 		    return;
 		}
-		if(strcmp("PTY", ptr) == 0) {
-#ifdef linux
-		    sprintf(devname, "/dev/%s", ptr);
-		    /* VERY linux dependent, recognize PTY as alias
-		       for all pseudo tty's */
-		    if((stat(devname, &stb) >= 0)
-		       && major(stb.st_rdev) == 4 
-		       && minor(stb.st_rdev) >= 192) {
-			fclose(f);
-			return;
-		    }
 #endif
+		if ((ptr[0] == '@') && isapty(tty) 
+		    && hnmatch(hostname, ptr+1)) {
+		    fclose(f);
+		    return;
 		}
 	    }
 	    /* if we get here, /etc/usertty exists, there's a line
@@ -969,12 +1082,32 @@ checktty(user, tty)
 	       name of the tty where the user is trying to log in.
 	       So deny access! */
 	    fclose(f);
-	    printf("Login on %s denied.\n", tty);
+	    printf("Login on %s from %s denied.\n", tty, hostname);
 	    badlogin(user);
 	    sleepexit(1);
 	}
     }
     fclose(f);
+
+    /* user is not explicitly mentioned in /etc/usertty, if there was
+       a default rule, use that */
+    if (defaultbuf[0]) {
+	strtok(defaultbuf, " \t");
+	while((ptr = strtok(NULL, "\t\n "))) {
+	    if (strncmp(tty, ptr, 10) == 0) return;
+#if 0
+	    if ((strcmp("PTY", ptr) == 0) && isapty(tty)) return;
+#endif
+	    if ((ptr[0] == '@') && isapty(tty) 
+		&& hnmatch(hostname, ptr+1)) return;
+	}
+
+	/* there was a default rule, but user didn't match, reject! */
+	printf("Login on %s from %s denied by default.\n", tty, hostname);
+	badlogin(user);
+	sleepexit(1);
+    }
+
     /* users not mentioned in /etc/usertty are by default allowed access
        on all tty's */
 }
